#include <stdio.h>
#include <math.h>
#include <errno.h>

#define MAXDATA 100  //change this value to change the max # of points

/*
 This Unix C tool calculates the linear regression with 1 or 2 parameters,
 y = a x  or  y = a x + b, respectively, for a set of points where both
 x and y values are affected by errors. The current implementation supports
 a maximum of MAXDATA data points.
 
 This source code should be named "linregr.c" and shoud be compiled with the following instruction:
 
 c linregr.c -o linregr
 
 The program is called with 2 mandatory and 2 optional arguments:
 
 linregr #par  filename  [optionally: conv]
 
 #par:			# of fitted parameters 1 (a) or 2 (a & b)
 filename		path to the input data file
 verb			verbosity:	0 = prints results only
							1 = prints input data and results (default)
							2 = prints also fitted function values with errors at data points (see NOTE below)
							3 = prints also convergence history
 conv			custom value for convergence goal - if omitted, the default is assumed (see below)
 
 The input file should consists of lines, one for each data point,
 containing 4 floating point values each, namely (in order)
 x-value	x-error		y-value		y-error
 
 NOTE: the error associated to the value of the fitted funcion takes into account only the incertainties (correlation matrix)
 of the parameters. In other words, it's the error on the function value calculated at points that coincide numerically
 with the data points, but are to be considered known arbitrary values, without error.
 
 Copyright 2008 A. Perego - University of Florence, INFN.
 
 USER LICENCE AGREEMENT
 
 This code is released without any explicit or implicit warranty regarding its exactness or its utility.
 The author takes no liability for any direct or indirect damage that might be caused by the use of this code.
 This code can be freely used, modified  and re-distributed both in the original or the modified form, with the
 following limitations:
 1- This user licence must be passed along with the code without changes.
 2- Modified copies of the code must be clearly marked as such.
 3- IT IS STRICLY FORBIDDEN TO  STORE, COMPILE, USE, MODIFY THIS CODE ON THE INFAMOUS "WINDOWS" PLATFORM, AS WELL AS UNDER THE CONTROL OF OR
	IN CONJUNCTION WITH ANY MICROSOFT SW/OS. Besides the risks of legal prosecution, whoever will not comply with this rule 
	will acquire TONS OF THE FILTHIEST KARMA and incur in a SPECIAL CURSE the author has taken from an ancient Egyptian manuscript,
	adapted and boosted for the occasion and unleashed upon the minions of the muddy side that should dare to infringe this rule.
*/
 
extern int linearregression(int npar, double * xa,double * dxa,double * ya,double * dya,int nd, double conv, int history, double * results);

int main (int argc, const char * argv[]) {
	FILE * myfile;
	int nd,ret,npar,iverb,ist,i;
	double x,dx,y,dy,conv,f,df;
	double xa[MAXDATA],ya[MAXDATA],dxa[MAXDATA],dya[MAXDATA],results[8];
	
	if(argc<3){
		printf("linregr usage: linregr #par input filename [optional: verbosity] [optional: convergence goal]\n");
		return -1;
	}
	sscanf(argv[1],"%d",&npar);
	if(argc>=4){
		sscanf(argv[3],"%d",&iverb);
	}else iverb=1;
	if(argc>=5){
		sscanf(argv[4],"%lg",&conv);
	}else conv=0.;
	nd=0;
	myfile=fopen(argv[2], "r");
	if(!myfile){
		printf("Error %d openind input file: %s\n",errno,strerror(errno));
		return -1;
	}
	while(fscanf(myfile,"%lg%lg%lg%lg",&x,&dx,&y,&dy)!=EOF ){
		if(nd>=MAXDATA){
			printf("# of input points exceeds current maximum (%d)\n",MAXDATA);
			return -1;
		}
		printf("x = %12g   dx = %12g   y = %12g   dy = %12g\n",x,dx,y,dy);
		xa[nd]=x;
		dxa[nd]=dx;
		ya[nd]=y;
		dya[nd++]=dy;
	}

	ist=(iverb> 2 ? 1 : 0);
	ret=linearregression(npar,xa,dxa,ya,dya,nd,conv,ist,results);
	if(!ret){
		if(npar<=1){
			printf("\ny=a x ; a = %g   Err(a) = %g   chi2 = %g  # of iter. %g.0\n",results[0],
				  results[2],results[5],results[7]);
		}else{
			printf("\ny=a x + b; a = %g   b = %g   Err(a) = %g  Err(b) = %g  cov = %g, chi2 = %g   # of iter. %g.0\n",results[0],
				   results[1],results[2],results[3],results[4],results[5],results[7]);
		}
		if(iverb>=2){
			printf("\nFitted function at data points\n\n");
			for(i=0;i<nd;i++){
				if(npar<2){
					f= results[0] * xa[i];
					df = fabs(xa[i])*results[2];
				}else{
					f = results[0] * xa[i] + results[1];
					df = sqrt(xa[i]*xa[i]*results[2]*results[2]+2.*xa[i]*results[4]+results[3]*results[3]);
				}
				printf("x=  %12g   y[exper]= %12g   fit= %12g   +-  %12g\n",xa[i],ya[i],f,df);
			}
		}
	}else{
		printf("linear fit failed: after %d iterations lambda = %g\n",results[7],results[6]);
	} 
    return 0;
}


int linearregression(int npar, double * xa, double * dxa, double * ya ,double * dya, int nd, double conv, int history, double * results)
/*
 this routine calculates a least squares fit on a straight line, y= a x + b, or y = a x 
 Its parameters are:
	npar:       if <= 1 y = a x  is fitted, if > 1 y = a x + b is fitted
	xa:			pointer to x values
	dxa:		pointier to errors on x values
	ya:			pointer to y values
	dya:		pointer to errors on y values
	nd:			number of data points
	conv:		if > 0. sets the convergence goal, if <= 0. the default value (1.e-4) is used
	history:	if > 0 prints lambda and chi2 at each iteration pass
	results:	pointer to array where results will be stored:
					results[0] = a
					results[1] = b			meaningful only when 2 parameters are fitted
					results[2] = Err(a)
					results[3] = Err(b)		meaningful only when 2 parameters are fitted
					results[4] = Cov(a,b)	meaningful only when 2 parameters are fitted
					results[5] = Chi^2
					results[6] = flamb
					results[7] = # of iterations performed

 the routine returns 0 upon normal completion, -1 when the fit didn't converge after 200 iterations
 */

 {
	double meanX[MAXDATA];  // a maximum of 100 data points
	double meanInc[MAXDATA];
	double a,b,tha,thb,z11,z12,z22,w1,w2,sx,sy,del,oldchi2,newchi2,yd,stot,flamb,xm,xminc,myconv;//nwx,nwy;
		int i,it;
		
		myconv=(conv > 0. ? conv : 1.e-4); //sets conv. goal to custom or default value
		a=(ya[nd-1]-ya[0])/(xa[nd-1]-xa[0]);  // guesses initial values for the parameters
		b= (npar > 1 ? ya[0]- a * xa[0] : 0.);
		for(i=0;i<nd;i++)meanX[i]=xa[i]; // initial values of x_tilde
		oldchi2=10.e30;   // a very large initial value
		flamb=0.001;      // usual initial value for L-M coefficient
		thb =0.;		// will stay at zero for 1-parameter fit
		for(it=0;it<200;it++){    // 200 iteration at maximum
			z11=z12=z22=w1=w2=0.;
			for(i=0;i<nd;i++){
				sx=dxa[i]*dxa[i];
				sy=dya[i]*dya[i];
				yd=ya[i]-a*xa[i]-b;
				stot=sy+a*a*sx;
				z11+=meanX[i]*meanX[i]/stot;
				w1+=yd*meanX[i]/stot;
				if(npar>1){
					z12+=meanX[i]/stot;
					z22+=1./stot;
					w2+=yd/stot;					
				}
			}
			z11*=1+flamb;		// solves 2 x 2 system with Cramer rule or calculates tha direclty
			if(npar>1){
				z22*=1+flamb;
				del=z11*z22-z12*z12;
				tha=(w1*z22-w2*z12)/del;
				thb=(w2*z11-w1*z12)/del;				
			}else{
				tha=w1/z11;
			}
			for(i=0;i<nd;i++){			// calculates increments of x_tilde_i
				sx=dxa[i]*dxa[i];
				sy=dya[i]*dya[i];
				xm=meanX[i];
				yd=ya[i]-(a+tha)*xm-b-thb;
				stot=sy+a*a*sx;
				xminc=(sy*(xa[i]-xm)+sx*a*yd)/(stot *(1.+flamb)); //increments must also be normalized to lambda factor
				xm+=xminc;
				meanInc[i]=xminc;	
				meanX[i]=xm;
			}
			a+=tha;
			b+=thb;
			newchi2=0.;   // caldulates new chi2
				for(i=0;i<nd;i++){
					sx=dxa[i]*dxa[i];
					sy=dya[i]*dya[i];
					xm=meanX[i];
					yd=ya[i]-a*meanX[i]-b;
					newchi2+= (xa[i]-meanX[i])*(xa[i]-meanX[i])/sx+ yd*yd/sy;				
				}
				if(history>0)printf("\t\t # iter = %3d   lambda = %12g   oldchi2 = %12g   newchi2 = %12g   rel chg. = %12g\n",
									it,flamb,oldchi2,newchi2,(newchi2-oldchi2)/oldchi2);
				if(newchi2>oldchi2){  // divergent step: multipl. lambda by 10. or reinit. to 0.001
					if(flamb>0){
						flamb*=10.;
					}else flamb=0.001;			
					a-=tha;   // restores old values of parameters and x_tilde_i
					b-=thb;
					for(i=0;i<nd;i++){
							meanX[i]-=meanInc[i]; 
					}
					continue;
				}
				if(fabs(newchi2-oldchi2)/oldchi2<myconv){  // convergence goal
					if(flamb>0.){ // stars a last step without L-M coeff.
						flamb=0.;
					}else break; // fit done
				}			
				oldchi2=newchi2;
				flamb/=10.;
		}	
		
		results[0]=a;
		results[1]=b;  // 0 for 1-par fit
	 if(npar>1){
		 results[2]=sqrt(z22/del);
		 results[3]=sqrt(z11/del);
		 results[4]=-z12/del;		 
	 }else{
		 results[2]=sqrt(1/z11);
	 }
		results[5]=newchi2;
		results[6]=flamb;
		results[7]=it;
			return (it<200 && flamb== 0. ? 0 : -1);
} 